/* GNU/Linux/MIPS specific low level interface, for the remote server for GDB.
   Copyright (C) 1995, 1996, 1998, 1999, 2000, 2001, 2002, 2005, 2006
   Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include "server.h"
#include "fbsd-low.h"
#include <sys/types.h>

#ifdef HAVE_SYS_REG_H
#include <sys/reg.h>
#else
#include <machine/reg.h>
#endif

/* This module only supports access to the general purpose registers.  */

#define mips_num_regs NUMSAVEREGS

static int mips_regmap[] = {
  ZERO * 4, AST * 4, V0 * 4, V1 * 4, A0 * 4, A1 * 4, A2 * 4, A3 * 4,
  T0   * 4, T1  * 4, T2 * 4, T3 * 4, TA0 * 4, TA1 * 4, TA2 * 4, TA3 * 4,
  S0   * 4, S1  * 4, S2 * 4, S3 * 4, S4 * 4, S5 * 4, S6 * 4, S7 * 4,
  T8   * 4, T9  * 4, K0 * 4, K1 * 4, GP * 4, SP * 4, S8 * 4, RA * 4,
  -1, MULLO * 4, MULHI * 4, BADVADDR * 4, CAUSE * 4, PC * 4
};


static int
mips_cannot_fetch_register (int regno)
{
  if (mips_regmap[regno] == -1)
    return 1;

  if (find_regno ("zero") == regno)
    return 1;

  return 0;
}

static int
mips_cannot_store_register (int regno)
{
  if (mips_regmap[regno] == -1)
    return 1;

  if (find_regno ("zero") == regno)
    return 1;

  if (find_regno ("cause") == regno)
    return 1;

  if (find_regno ("bad") == regno)
    return 1;

  if (find_regno ("fir") == regno)
    return 1;

  return 0;
}

#include <sys/procfs.h>
#include <sys/ptrace.h>

static void
mips_fill_gregset (void *buf)
{
  int i;
  register_t val;
  char regbuf[16];

  for (i = 0; i <= PC; i++) 
    {
      collect_register (i, regbuf);
      val = (register_t) extract_signed_integer (regbuf, register_size(i));
      memcpy (((char *) buf) + mips_regmap[i], &val, sizeof (register_t));
    }
}

static void
mips_store_gregset (const void *buf)
{
  int i;
  long long val;
  char regbuf[16];

  for (i = 0; i <= PC; i++) 
    {
      val = extract_signed_integer (((char *)buf) + mips_regmap[i], 
				    sizeof (register_t));
      store_signed_integer (regbuf, register_size(i), val);
      supply_register (i, regbuf);
    }
}

struct regset_info target_regsets[] = {
  { PT_GETREGS, PT_SETREGS, sizeof (struct reg),
    GENERAL_REGS,
    mips_fill_gregset, mips_store_gregset },
  { 0, 0, -1, -1, NULL, NULL }
};


static CORE_ADDR
mips_get_pc ()
{
  unsigned long pc;
  collect_register_by_name ("pc", &pc);
  return pc;
}

static void
mips_set_pc (CORE_ADDR pc)
{
  unsigned long newpc = pc;
  supply_register_by_name ("pc", &newpc);
}

/* Correct in either endianness.  */
static const unsigned long mips_breakpoint = 0x0005000d;
#define mips_breakpoint_len 4

/* We only place breakpoints in empty marker functions, and thread locking
   is outside of the function.  So rather than importing software single-step,
   we can just run until exit.  */
static CORE_ADDR
mips_reinsert_addr ()
{
  unsigned long pc;
  collect_register_by_name ("ra", &pc);
  return pc;
}

static int
mips_breakpoint_at (CORE_ADDR where)
{
  unsigned long insn;

  (*the_target->read_memory) (where, (unsigned char *) &insn, 4);
  if (insn == mips_breakpoint)
    return 1;

  /* If necessary, recognize more trap instructions here.  GDB only uses the
     one.  */
  return 0;
}

struct fbsd_target_ops the_low_target = {
  mips_num_regs,
  mips_regmap,
  mips_cannot_fetch_register,
  mips_cannot_store_register,
  mips_get_pc,
  mips_set_pc,
  (const unsigned char *) &mips_breakpoint,
  mips_breakpoint_len,
  mips_reinsert_addr,
  0,
  mips_breakpoint_at,
};
