/*-
 * Copyright (c) 2006 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/gnu/usr.bin/gdb/kgdb/trgt_powerpc.c,v 1.2 2006/10/16 20:07:23 jhb Exp $");

#include <sys/types.h>
#include <err.h>
#include <kvm.h>
#include <string.h>

#include <defs.h>
#include <target.h>
#include <gdbthread.h>
#include <inferior.h>
#include <regcache.h>
#include <frame-unwind.h>
#include <objfiles.h>
#include <ppc-tdep.h>

#include "kgdb.h"

static size_t tf_fixreg_off, tf_lr_off, tf_cr_off, tf_xer_off, 
	tf_ctr_off, tf_srr0_off;
static size_t pcb_size, pcb_context_off, pcb_sp_off, pcb_lr_off, pcb_cr_off;
static CORE_ADDR kernel_toc_addr;

static CORE_ADDR trap_start = (CORE_ADDR) -1, trap_end = (CORE_ADDR) -1;

void
kgdb_trgt_fetch_registers(int regno __unused)
{
	static CORE_ADDR kernel_toc;
	struct kthr *kt;
	struct gdbarch_tdep *tdep;
	int i, regsz;
	char* pcb_ptr = alloca(pcb_size);

	tdep = gdbarch_tdep (current_gdbarch);
	regsz = register_size(current_gdbarch, PC_REGNUM);

	kt = kgdb_thr_lookup_tid(ptid_get_pid(inferior_ptid));
	if (kt == NULL)
		return;
	if (kvm_read(kvm, kt->pcb, pcb_ptr, pcb_size) != pcb_size) {
		warnx("kvm_read: %s", kvm_geterr(kvm));
		memset(pcb_ptr, 0, pcb_size);
	}

	/* Read the TOC value if the address is set */
	if (kernel_toc_addr && kernel_toc == 0) {
		kvm_read(kvm, kernel_toc_addr, &kernel_toc, sizeof(kernel_toc));
	}

	/*
	 * r12-r31 are saved in the pcb
	 */
	for (i = 12; i <= 31; i++) {
		regcache_raw_supply (current_regcache, tdep->ppc_gp0_regnum + i,
			     pcb_ptr + pcb_context_off + (i - 12) * regsz);
	}

	/* r1 is saved in the sp field */
	regcache_raw_supply (current_regcache, tdep->ppc_gp0_regnum + 1, 
			     pcb_ptr + pcb_sp_off);
	/* TOC value is supplied as the value of r2 */
	regcache_raw_supply (current_regcache, tdep->ppc_gp0_regnum + 2, 
			     (char*) &kernel_toc);
	regcache_raw_supply (current_regcache, PC_REGNUM, pcb_ptr + pcb_lr_off);
	regcache_raw_supply (current_regcache, tdep->ppc_lr_regnum,
			     pcb_ptr + pcb_lr_off);
	regcache_raw_supply (current_regcache, tdep->ppc_cr_regnum,
			     pcb_ptr + pcb_cr_off);
}

void
kgdb_trgt_store_registers(int regno __unused)
{
	fprintf_unfiltered(gdb_stderr, "XXX: %s\n", __func__);
}

void
kgdb_trgt_init_type_info(void)
{
	struct minimal_symbol *sym;
	struct gdbarch_tdep *tdep;
	char buf[50];

	sym = lookup_minimal_symbol ("interrupt_vector_base", NULL, NULL);
	if (sym)
		trap_start = SYMBOL_VALUE_ADDRESS (sym);
	else
		trap_start = (CORE_ADDR) -1;

	sym = lookup_minimal_symbol ("interrupt_vector_end", NULL, NULL);
	if (sym)
		trap_end = SYMBOL_VALUE_ADDRESS (sym);
	else
		trap_end = (CORE_ADDR) -1;

	tf_fixreg_off = kgdb_parse("&((struct trapframe *)0)->fixreg[0]");
	tf_lr_off = kgdb_parse("&((struct trapframe *)0)->lr");
	tf_cr_off = kgdb_parse("&((struct trapframe *)0)->cr");
	tf_xer_off = kgdb_parse("&((struct trapframe *)0)->xer");
	tf_ctr_off = kgdb_parse("&((struct trapframe *)0)->ctr");
	tf_srr0_off = kgdb_parse("&((struct trapframe *)0)->srr0");

	pcb_size = kgdb_parse("sizeof(struct pcb)");
	pcb_context_off = kgdb_parse("&((struct pcb *)0)->pcb_context[0]");
	pcb_sp_off = kgdb_parse("&((struct pcb *)0)->pcb_sp");
	pcb_lr_off = kgdb_parse("&((struct pcb *)0)->pcb_lr");
	pcb_cr_off = kgdb_parse("&((struct pcb *)0)->pcb_cr");

	/* Get the address of the kernel TOC, if available. Too early
	   to read the actual value since kvm has not been initialized yet. */
	sym = lookup_minimal_symbol ("__pcpu", NULL, NULL);
	if (sym) {
		/* TOC may or may not be defined in struct pcpu. 
		   Set the address of TOC only if it's available*/
		errno = 0;
		size_t toc_off = 
			kgdb_parse("&((struct pcpu *)0)->pc_kernel_toc");
		kernel_toc_addr = errno ? 
			0 : SYMBOL_VALUE_ADDRESS (sym) + toc_off;
	}
}

struct kgdb_frame_cache {
	CORE_ADDR	pc;
	CORE_ADDR	sp;
};

static struct kgdb_frame_cache *
kgdb_trgt_frame_cache(struct frame_info *next_frame, void **this_cache)
{
	char buf[MAX_REGISTER_SIZE];
	struct kgdb_frame_cache *cache;

	cache = *this_cache;
	if (cache == NULL) {
		cache = FRAME_OBSTACK_ZALLOC(struct kgdb_frame_cache);
		*this_cache = cache;
		cache->pc = frame_func_unwind(next_frame);
		frame_unwind_register(next_frame, SP_REGNUM, buf);
		cache->sp = extract_unsigned_integer(buf,
		    register_size(current_gdbarch, SP_REGNUM));
	}
	return (cache);
}

static void
kgdb_trgt_trapframe_this_id(struct frame_info *next_frame, void **this_cache,
    struct frame_id *this_id)
{
	struct kgdb_frame_cache *cache;

	cache = kgdb_trgt_frame_cache(next_frame, this_cache);
	*this_id = frame_id_build(cache->sp, cache->pc);
}

static void
kgdb_trgt_trapframe_prev_register(struct frame_info *next_frame,
    void **this_cache, int regnum, int *optimizedp, enum lval_type *lvalp,
    CORE_ADDR *addrp, int *realnump, void *valuep)
{
	char dummy_valuep[MAX_REGISTER_SIZE];
	struct gdbarch_tdep *tdep;
	struct kgdb_frame_cache *cache;
	int ofs, regsz;

	tdep = gdbarch_tdep(current_gdbarch);
	regsz = register_size(current_gdbarch, regnum);

	if (valuep == NULL)
		valuep = dummy_valuep;
	memset(valuep, 0, regsz);
	*optimizedp = 0;
	*addrp = 0;
	*lvalp = not_lval;
	*realnump = -1;

	if (regnum >= tdep->ppc_gp0_regnum &&
	    regnum < tdep->ppc_gp0_regnum + ppc_num_gprs)
		ofs = tf_fixreg_off + (regnum - tdep->ppc_gp0_regnum) * regsz;
	else if (regnum == tdep->ppc_lr_regnum)
		ofs = tf_lr_off;
	else if (regnum == tdep->ppc_cr_regnum)
		ofs = tf_cr_off;
	else if (regnum == tdep->ppc_xer_regnum)
		ofs = tf_xer_off;
	else if (regnum == tdep->ppc_ctr_regnum)
		ofs = tf_ctr_off;
	else if (regnum == PC_REGNUM)
		ofs = tf_srr0_off;
	else
		return;

	cache = kgdb_trgt_frame_cache(next_frame, this_cache);
	*addrp = cache->sp + 8 + ofs;
	*lvalp = lval_memory;
	target_read_memory(*addrp, valuep, regsz);
}

static const struct frame_unwind kgdb_trgt_trapframe_unwind = {
	NORMAL_FRAME,
	&kgdb_trgt_trapframe_this_id,
	&kgdb_trgt_trapframe_prev_register,
	NULL,
	&kgdb_trgt_trapframe_sniffer,
	NULL
};

const struct frame_unwind *
kgdb_trgt_trapframe_sniffer(struct frame_info *next_frame)
{
	char *pname;
	CORE_ADDR pc;

	pc = frame_pc_unwind(next_frame);

	if (trap_start == (CORE_ADDR) -1 || trap_end == (CORE_ADDR) -1) {
		pname = NULL;
		find_pc_partial_function(pc, &pname, NULL, NULL);
		if (pname == NULL)
			return (NULL);

		if (strcmp(pname, "asttrapexit") == 0 ||
		    strcmp(pname, "trapexit") == 0)
		    return (&kgdb_trgt_trapframe_unwind);
	}
	else {
		if (pc >= trap_start && pc < trap_end)
		    return (&kgdb_trgt_trapframe_unwind);
	}

	return (NULL);
}
